From bc84474327c6b0ebb31975ee7c99dbbce257166c Mon Sep 17 00:00:00 2001
From: Meowthink <meowthink@googlemail.com>
Date: Tue, 31 Jul 2012 17:35:02 +0800
Subject: [PATCH] Add 64-bit kernel / 32-bit world compatible wrappers to
 drm2/i915kms

---
 sys/dev/drm2/drmP.h               |   4 +
 sys/dev/drm2/drm_drv.c            |  24 ++
 sys/dev/drm2/drm_ioc32.c          | 781 ++++++++++++++++++++++++++++++++++++++
 sys/dev/drm2/i915/i915_dma.c      |   7 +
 sys/dev/drm2/i915/i915_ioc32.c    | 124 ++++++
 sys/modules/drm2/drm2/Makefile    |   2 +-
 sys/modules/drm2/i915kms/Makefile |   3 +-
 7 files changed, 943 insertions(+), 2 deletions(-)
 create mode 100644 sys/dev/drm2/drm_ioc32.c
 create mode 100644 sys/dev/drm2/i915/i915_ioc32.c

diff --git a/sys/dev/drm2/drmP.h b/sys/dev/drm2/drmP.h
index 16227b5..c3cbacb 100644
--- a/sys/dev/drm2/drmP.h
+++ b/sys/dev/drm2/drmP.h
@@ -105,6 +105,7 @@ struct drm_file;
 #include <dev/drm2/drm_mm.h>
 #include <dev/drm2/drm_hashtab.h>
 
+#include "opt_compat.h"
 #include "opt_drm.h"
 #ifdef DRM_DEBUG
 #undef DRM_DEBUG
@@ -760,6 +761,9 @@ struct drm_driver_info {
 	int	(*device_is_agp) (struct drm_device * dev);
 
 	drm_ioctl_desc_t *ioctls;
+#ifdef COMPAT_FREEBSD32
+	drm_ioctl_desc_t *compat_ioctls;
+#endif
 	int	max_ioctl;
 
 	int	buf_priv_size;
diff --git a/sys/dev/drm2/drm_drv.c b/sys/dev/drm2/drm_drv.c
index 42ff194..ab4c850 100644
--- a/sys/dev/drm2/drm_drv.c
+++ b/sys/dev/drm2/drm_drv.c
@@ -179,6 +179,11 @@ static drm_ioctl_desc_t		  drm_ioctls[256] = {
 	DRM_IOCTL_DEF(DRM_IOCTL_MODE_DESTROY_DUMB, drm_mode_destroy_dumb_ioctl, DRM_MASTER|DRM_CONTROL_ALLOW|DRM_UNLOCKED),
 };
 
+#ifdef COMPAT_FREEBSD32
+#include "drm_ioc32.c"
+#include <sys/sysent.h>
+#endif
+
 static struct cdevsw drm_cdevsw = {
 	.d_version =	D_VERSION,
 	.d_open =	drm_open,
@@ -846,6 +851,19 @@ int drm_ioctl(struct cdev *kdev, u_long cmd, caddr_t data, int flags,
 		return EINVAL;
 	}
 
+#ifdef COMPAT_FREEBSD32
+	/* Called whenever a 32-bit process running under a 64-bit kernel
+	 * performs an ioctl on /dev/drm.
+	 */
+	if (SV_CURPROC_FLAG(SV_ILP32)
+		&& drm_compat_ioctls[nr].func != NULL)
+		/* Assume that ioctls without an explicit compat routine will just
+		 * work.  This may not always be a good assumption, but it's better
+		 * than always failing.
+		 */
+		ioctl = &drm_compat_ioctls[nr];
+	else
+#endif
 	ioctl = &drm_ioctls[nr];
 	/* It's not a core DRM ioctl, try driver-specific. */
 	if (ioctl->func == NULL && nr >= DRM_COMMAND_BASE) {
@@ -856,6 +874,12 @@ int drm_ioctl(struct cdev *kdev, u_long cmd, caddr_t data, int flags,
 			    nr, dev->driver->max_ioctl);
 			return EINVAL;
 		}
+#ifdef COMPAT_FREEBSD32
+		if (SV_CURPROC_FLAG(SV_ILP32)
+			&& dev->driver->compat_ioctls[nr].func != NULL)
+			ioctl = &dev->driver->compat_ioctls[nr];
+		else
+#endif
 		ioctl = &dev->driver->ioctls[nr];
 		is_driver_ioctl = 1;
 	}
diff --git a/sys/dev/drm2/drm_ioc32.c b/sys/dev/drm2/drm_ioc32.c
new file mode 100644
index 0000000..906116b
--- /dev/null
+++ b/sys/dev/drm2/drm_ioc32.c
@@ -0,0 +1,781 @@
+/*-
+ * Copyright (C) Paul Mackerras 2005.
+ * All Rights Reserved.
+ *
+ * Permission is hereby granted, free of charge, to any person obtaining a
+ * copy of this software and associated documentation files (the "Software"),
+ * to deal in the Software without restriction, including without limitation
+ * the rights to use, copy, modify, merge, publish, distribute, sublicense,
+ * and/or sell copies of the Software, and to permit persons to whom the
+ * Software is furnished to do so, subject to the following conditions:
+ *
+ * The above copyright notice and this permission notice (including the next
+ * paragraph) shall be included in all copies or substantial portions of the
+ * Software.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+ * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
+ * THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
+ * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+ * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
+ * IN THE SOFTWARE.
+ *
+ * Authors:
+ *    Paul Mackerras <paulus@samba.org>
+ */
+
+__FBSDID("$FreeBSD$");
+
+/** @file drm_ioc32.c
+ * 32-bit ioctl compatibility routines for the DRM.
+ */
+ 
+#include <sys/types.h>
+#include <sys/malloc.h>
+
+#define DRM_IOCTL_VERSION32		DRM_IOWR(0x00, drm_version32_t)
+#define DRM_IOCTL_GET_UNIQUE32		DRM_IOWR(0x01, drm_unique32_t)
+#define DRM_IOCTL_GET_MAP32		DRM_IOWR(0x04, drm_map32_t)
+#define DRM_IOCTL_GET_CLIENT32		DRM_IOWR(0x05, drm_client32_t)
+#define DRM_IOCTL_GET_STATS32		DRM_IOR( 0x06, drm_stats32_t)
+
+#define DRM_IOCTL_SET_UNIQUE32		DRM_IOW( 0x10, drm_unique32_t)
+#define DRM_IOCTL_ADD_MAP32		DRM_IOWR(0x15, drm_map32_t)
+#define DRM_IOCTL_ADD_BUFS32		DRM_IOWR(0x16, drm_buf_desc32_t)
+#define DRM_IOCTL_MARK_BUFS32		DRM_IOW( 0x17, drm_buf_desc32_t)
+#define DRM_IOCTL_INFO_BUFS32		DRM_IOWR(0x18, drm_buf_info32_t)
+#define DRM_IOCTL_MAP_BUFS32		DRM_IOWR(0x19, drm_buf_map32_t)
+#define DRM_IOCTL_FREE_BUFS32		DRM_IOW( 0x1a, drm_buf_free32_t)
+
+#define DRM_IOCTL_RM_MAP32		DRM_IOW( 0x1b, drm_map32_t)
+
+#define DRM_IOCTL_SET_SAREA_CTX32	DRM_IOW( 0x1c, drm_ctx_priv_map32_t)
+#define DRM_IOCTL_GET_SAREA_CTX32	DRM_IOWR(0x1d, drm_ctx_priv_map32_t)
+
+#define DRM_IOCTL_RES_CTX32		DRM_IOWR(0x26, drm_ctx_res32_t)
+#define DRM_IOCTL_DMA32			DRM_IOWR(0x29, drm_dma32_t)
+
+#define DRM_IOCTL_AGP_ENABLE32		DRM_IOW( 0x32, drm_agp_mode32_t)
+#define DRM_IOCTL_AGP_INFO32		DRM_IOR( 0x33, drm_agp_info32_t)
+#define DRM_IOCTL_AGP_ALLOC32		DRM_IOWR(0x34, drm_agp_buffer32_t)
+#define DRM_IOCTL_AGP_FREE32		DRM_IOW( 0x35, drm_agp_buffer32_t)
+#define DRM_IOCTL_AGP_BIND32		DRM_IOW( 0x36, drm_agp_binding32_t)
+#define DRM_IOCTL_AGP_UNBIND32		DRM_IOW( 0x37, drm_agp_binding32_t)
+
+#define DRM_IOCTL_SG_ALLOC32		DRM_IOW( 0x38, drm_scatter_gather32_t)
+#define DRM_IOCTL_SG_FREE32		DRM_IOW( 0x39, drm_scatter_gather32_t)
+
+#define DRM_IOCTL_UPDATE_DRAW32		DRM_IOW( 0x3f, drm_update_draw32_t)
+
+#define DRM_IOCTL_WAIT_VBLANK32		DRM_IOWR(0x3a, drm_wait_vblank32_t)
+
+typedef struct drm_version_32 {
+	int version_major;	  /**< Major version */
+	int version_minor;	  /**< Minor version */
+	int version_patchlevel;	   /**< Patch level */
+	u32 name_len;		  /**< Length of name buffer */
+	u32 name;		  /**< Name of driver */
+	u32 date_len;		  /**< Length of date buffer */
+	u32 date;		  /**< User-space buffer to hold date */
+	u32 desc_len;		  /**< Length of desc buffer */
+	u32 desc;		  /**< User-space buffer to hold desc */
+} drm_version32_t;
+
+static int compat_drm_version(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_version32_t *v32 = data;
+	struct drm_version version;
+	int err;
+
+	version.name_len = v32->name_len;
+	version.name = (void *)(unsigned long)v32->name;
+	version.date_len = v32->date_len;
+	version.date = (void *)(unsigned long)v32->date;
+	version.desc_len = v32->desc_len;
+	version.desc = (void *)(unsigned long)v32->desc;
+	
+	err = drm_version(dev, (void *)&version, file_priv);
+	if (err)
+		return err;
+
+	v32->version_major = version.version_major;
+	v32->version_minor = version.version_minor;
+	v32->version_patchlevel = version.version_patchlevel;
+	v32->name_len = version.name_len;
+	v32->date_len = version.date_len;
+	v32->desc_len = version.desc_len;
+
+	return 0;
+}
+
+typedef struct drm_unique32 {
+	u32 unique_len;	/**< Length of unique */
+	u32 unique;	/**< Unique name for driver instantiation */
+} drm_unique32_t;
+
+static int compat_drm_getunique(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_unique32_t *uq32 = data;
+	struct drm_unique u;
+	int err;
+
+	u.unique_len = uq32->unique_len;
+	u.unique = (void *)(unsigned long)uq32->unique;
+
+	err = drm_getunique(dev, (void *)&u, file_priv);
+	if (err)
+		return err;
+
+	uq32->unique_len = u.unique_len;
+
+	return 0;
+}
+
+static int compat_drm_setunique(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_unique32_t *uq32 = data;
+	struct drm_unique u;
+
+	u.unique_len = uq32->unique_len;
+	u.unique = (void *)(unsigned long)uq32->unique;
+
+	return drm_setunique(dev, (void *)&u, file_priv);
+}
+
+typedef struct drm_map32 {
+	u32 offset;		/**< Requested physical address (0 for SAREA)*/
+	u32 size;		/**< Requested physical size (bytes) */
+	enum drm_map_type type;	/**< Type of memory to map */
+	enum drm_map_flags flags;	/**< Flags */
+	u32 handle;		/**< User-space: "Handle" to pass to mmap() */
+	int mtrr;		/**< MTRR slot used */
+} drm_map32_t;
+
+static int compat_drm_getmap(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_map32_t *m32 = data;
+	struct drm_map map;
+	int err;
+	void *handle;
+
+	map.offset = (unsigned long)m32->offset;
+
+	err = drm_getmap(dev, (void *)&map, file_priv);
+	if (err)
+		return err;
+
+	m32->offset = map.offset;
+	m32->size = map.size;
+	m32->type = map.type;
+	m32->flags = map.flags;
+	handle = map.handle;
+	m32->mtrr = map.mtrr;
+
+	m32->handle = (unsigned long)handle;
+
+	return 0;
+
+}
+
+static int compat_drm_addmap(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_map32_t *m32 = data;
+	struct drm_map map;
+	int err;
+	void *handle;
+	
+	map.offset = (unsigned long)m32->offset;
+	map.size = (unsigned long)m32->size;
+	map.type = m32->type;
+	map.flags = m32->flags;
+
+	err = drm_addmap_ioctl(dev, (void *)&map, file_priv);
+	if (err)
+		return err;
+
+	m32->offset = map.offset;
+	m32->mtrr = map.mtrr;
+	handle = map.handle;
+	
+	m32->handle = (unsigned long)handle;
+	if (m32->handle != (unsigned long)handle)
+		DRM_DEBUG("compat_drm_addmap truncated handle"
+				   " %p for type %d offset %x\n",
+				   handle, m32->type, m32->offset);
+
+	return 0;
+}
+
+static int compat_drm_rmmap(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_map32_t *m32 = data;
+	struct drm_map map;
+
+	map.handle = (void *)(unsigned long)m32->handle;
+
+	return drm_rmmap_ioctl(dev, (void *)&map, file_priv);
+}
+
+typedef struct drm_client32 {
+	int idx;	/**< Which client desired? */
+	int auth;	/**< Is client authenticated? */
+	u32 pid;	/**< Process ID */
+	u32 uid;	/**< User ID */
+	u32 magic;	/**< Magic */
+	u32 iocs;	/**< Ioctl count */
+} drm_client32_t;
+
+static int compat_drm_getclient(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_client32_t *c32 = data;
+	struct drm_client client;
+	int err;
+
+	client.idx = c32->idx;
+
+	err = drm_getclient(dev, (void *)&client, file_priv);
+	if (err)
+		return err;
+
+	c32->idx = client.idx;
+	c32->auth = client.auth;
+	c32->pid = client.pid;
+	c32->uid = client.uid;
+	c32->magic = client.magic;
+	c32->iocs = client.iocs;
+
+	return 0;
+}
+
+typedef struct drm_stats32 {
+	u32 count;
+	struct {
+		u32 value;
+		enum drm_stat_type type;
+	} data[15];
+} drm_stats32_t;
+
+static int compat_drm_getstats(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_stats32_t *s32 = data;
+	struct drm_stats stats;
+	int i, err;
+
+	err = drm_getstats(dev, (void *)&stats, file_priv);
+	if (err)
+		return err;
+
+	s32->count = stats.count;
+	for (i = 0; i < stats.count; i++) {
+		s32->data[i].value = stats.data[i].value;
+		s32->data[i].type = stats.data[i].type;
+	}
+
+	return 0;
+}
+
+typedef struct drm_buf_desc32 {
+	int count;		 /**< Number of buffers of this size */
+	int size;		 /**< Size in bytes */
+	int low_mark;		 /**< Low water mark */
+	int high_mark;		 /**< High water mark */
+	int flags;
+	u32 agp_start;		 /**< Start address in the AGP aperture */
+} drm_buf_desc32_t;
+
+static int compat_drm_addbufs(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_buf_desc32_t *b32 = data;
+	struct drm_buf_desc buf;
+	int err;
+
+	buf.count = b32->count;
+	buf.size = b32->size;
+	buf.low_mark = b32->low_mark;
+	buf.high_mark = b32->high_mark;
+	buf.flags = b32->flags;
+	buf.agp_start = (unsigned long)b32->agp_start;
+
+	err = drm_addbufs(dev, (void *)&buf, file_priv);
+	if (err)
+		return err;
+
+	b32->count = buf.count;
+	b32->size = buf.size;
+	b32->low_mark = buf.low_mark;
+	b32->high_mark = buf.high_mark;
+	b32->flags = buf.flags;
+	b32->agp_start = buf.agp_start;
+	
+	return 0;
+}
+
+static int compat_drm_markbufs(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_buf_desc32_t *b32 = data;
+	struct drm_buf_desc buf;
+
+	buf.size = b32->size;
+	buf.low_mark = b32->low_mark;
+	buf.high_mark = b32->high_mark;
+	
+	return drm_markbufs(dev, (void *)&buf, file_priv);
+}
+
+typedef struct drm_buf_info32 {
+	int count;		/**< Entries in list */
+	u32 list;
+} drm_buf_info32_t;
+
+static int compat_drm_infobufs(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_buf_info32_t *req32 = data;
+	drm_buf_desc32_t *to;
+	struct drm_buf_info *request;
+	struct drm_buf_desc *list;
+	size_t nbytes;
+	int i, err;
+	int count, actual;
+
+	count = req32->count;
+	to = (drm_buf_desc32_t *)(unsigned long)req32->list;
+	if (count < 0)
+		count = 0;
+
+	nbytes = sizeof(*request) + count * sizeof(struct drm_buf_desc);
+	request = malloc(nbytes, DRM_MEM_BUFLISTS, M_ZERO | M_NOWAIT);
+	if (!request)
+		return -EFAULT;
+	list = (struct drm_buf_desc *) (request + 1);
+
+	request->count = count;
+	request->list = list;
+
+	err = drm_infobufs(dev, (void *)request, file_priv);
+	if (err)
+		return err;
+
+	actual = request->count;
+	if (count >= actual)
+		for (i = 0; i < actual; ++i) {
+			to[i].count = list[i].count;
+			to[i].size = list[i].size;
+			to[i].low_mark = list[i].low_mark;
+			to[i].high_mark = list[i].high_mark;
+			to[i].flags = list[i].flags;
+		}
+
+	req32->count = actual;
+
+	return 0;
+}
+
+typedef struct drm_buf_pub32 {
+	int idx;		/**< Index into the master buffer list */
+	int total;		/**< Buffer size */
+	int used;		/**< Amount of buffer in use (for DMA) */
+	u32 address;		/**< Address of buffer */
+} drm_buf_pub32_t;
+
+typedef struct drm_buf_map32 {
+	int count;		/**< Length of the buffer list */
+	u32 virtual;		/**< Mmap'd area in user-virtual */
+	u32 list;		/**< Buffer information */
+} drm_buf_map32_t;
+
+static int compat_drm_mapbufs(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_buf_map32_t *req32 = data;
+	drm_buf_pub32_t *list32;
+	struct drm_buf_map *request;
+	struct drm_buf_pub *list;
+	int i, err;
+	int count, actual;
+	size_t nbytes;
+
+	count = req32->count;
+	list32 = (void *)(unsigned long)req32->list;
+
+	if (count < 0)
+		return -EINVAL;
+	nbytes = sizeof(*request) + count * sizeof(struct drm_buf_pub);
+	request = malloc(nbytes, DRM_MEM_BUFLISTS, M_ZERO | M_NOWAIT);
+	if (!request)
+		return -EFAULT;
+	list = (struct drm_buf_pub *) (request + 1);
+
+	request->count = count;
+	request->list = list;
+
+	err = drm_mapbufs(dev, (void *)request, file_priv);
+	if (err)
+		return err;
+
+	actual = request->count;
+	if (count >= actual)
+		for (i = 0; i < actual; ++i) {
+			list32[i].idx = list[i].idx;
+			list32[i].total = list[i].total;
+			list32[i].used = list[i].used;
+			list32[i].address = (unsigned long)list[i].address;
+		}
+
+	req32->count = actual;
+	req32->virtual = (unsigned long)request->virtual;
+
+	return 0;
+}
+
+typedef struct drm_buf_free32 {
+	int count;
+	u32 list;
+} drm_buf_free32_t;
+
+static int compat_drm_freebufs(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_buf_free32_t *req32 = data;
+	struct drm_buf_free request;
+
+	request.count = req32->count;
+	request.list = (int *)(unsigned long)req32->list;
+
+	return drm_freebufs(dev, (void *)&request, file_priv);
+}
+
+typedef struct drm_ctx_priv_map32 {
+	unsigned int ctx_id;	 /**< Context requesting private mapping */
+	u32 handle;		/**< Handle of map */
+} drm_ctx_priv_map32_t;
+
+static int compat_drm_setsareactx(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_ctx_priv_map32_t *req32 = data;
+	struct drm_ctx_priv_map request;
+
+	request.ctx_id = req32->ctx_id;
+	request.handle = (void *)(unsigned long)req32->handle;
+
+	return drm_setsareactx(dev, (void *)&request, file_priv);
+}
+
+static int compat_drm_getsareactx(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_ctx_priv_map32_t *req32 = data;
+	struct drm_ctx_priv_map request;
+	int err;
+
+	request.ctx_id = req32->ctx_id;
+
+	err = drm_getsareactx(dev, (void *)&request, file_priv);
+	if (err)
+		return err;
+
+	req32->handle = (unsigned long)request.handle;
+
+	return 0;
+}
+
+typedef struct drm_ctx_res32 {
+	int count;
+	u32 contexts;
+} drm_ctx_res32_t;
+
+static int compat_drm_resctx(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_ctx_res32_t *res32 = data;
+	struct drm_ctx_res res;
+	int err;
+
+	res.count = res32->count;
+	res.contexts = (struct drm_ctx __user *)(unsigned long)res32->contexts;
+
+	err = drm_resctx(dev, (void *)&res, file_priv);
+	if (err)
+		return err;
+
+	res32->count = res.count;
+
+	return 0;
+}
+
+typedef struct drm_dma32 {
+	int context;		  /**< Context handle */
+	int send_count;		  /**< Number of buffers to send */
+	u32 send_indices;	  /**< List of handles to buffers */
+	u32 send_sizes;		  /**< Lengths of data to send */
+	enum drm_dma_flags flags;		  /**< Flags */
+	int request_count;	  /**< Number of buffers requested */
+	int request_size;	  /**< Desired size for buffers */
+	u32 request_indices;	  /**< Buffer information */
+	u32 request_sizes;
+	int granted_count;	  /**< Number of buffers granted */
+} drm_dma32_t;
+
+static int compat_drm_dma(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_dma32_t *d32 = data;
+	struct drm_dma d;
+	int err;
+
+	d.context = d32->context;
+	d.send_count = d32->send_count;
+	d.send_indices = (int *)(unsigned long)d32->send_indices;
+	d.send_sizes = (int *)(unsigned long)d32->send_sizes;
+	d.flags = d32->flags;
+	d.request_count = d32->request_count;
+	d.request_indices = (int *)(unsigned long)d32->request_indices;
+	d.request_sizes = (int *)(unsigned long)d32->request_sizes;
+
+	err = drm_dma(dev, (void *)&d, file_priv);
+	if (err)
+		return err;
+
+	d32->request_size = d.request_size;
+	d32->granted_count = d.granted_count;
+
+	return 0;
+}
+
+typedef struct drm_agp_mode32 {
+	u32 mode;	/**< AGP mode */
+} drm_agp_mode32_t;
+
+static int compat_drm_agp_enable(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_agp_mode32_t *m32 = data;
+	struct drm_agp_mode mode;
+
+	mode.mode = m32->mode;
+
+	return drm_agp_enable_ioctl(dev, (void *)&mode, file_priv);
+}
+
+typedef struct drm_agp_info32 {
+	int agp_version_major;
+	int agp_version_minor;
+	u32 mode;
+	u32 aperture_base;	/* physical address */
+	u32 aperture_size;	/* bytes */
+	u32 memory_allowed;	/* bytes */
+	u32 memory_used;
+
+	/* PCI information */
+	unsigned short id_vendor;
+	unsigned short id_device;
+} drm_agp_info32_t;
+
+static int compat_drm_agp_info(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_agp_info32_t *i32 = data;
+	struct drm_agp_info info;
+	int err;
+
+	err = drm_agp_info_ioctl(dev, (void *)&info, file_priv);
+	if (err)
+		return err;
+
+	i32->agp_version_major = info.agp_version_major;
+	i32->agp_version_minor = info.agp_version_minor;
+	i32->mode = info.mode;
+	i32->aperture_base = info.aperture_base;
+	i32->aperture_size = info.aperture_size;
+	i32->memory_allowed = info.memory_allowed;
+	i32->memory_used = info.memory_used;
+	i32->id_vendor = info.id_vendor;
+	i32->id_device = info.id_device;
+
+	return 0;
+}
+
+typedef struct drm_agp_buffer32 {
+	u32 size;	/**< In bytes -- will round to page boundary */
+	u32 handle;	/**< Used for binding / unbinding */
+	u32 type;	/**< Type of memory to allocate */
+	u32 physical;	/**< Physical used by i810 */
+} drm_agp_buffer32_t;
+
+static int compat_drm_agp_alloc(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_agp_buffer32_t *req32 = data;
+	struct drm_agp_buffer request;
+	int err;
+
+	request.size = req32->size;
+	request.type = req32->type;
+
+	err = drm_agp_alloc_ioctl(dev, (void *)&request, file_priv);
+	if (err)
+		return err;
+
+	req32->handle = request.handle;
+	req32->physical = request.physical;
+
+	return 0;
+}
+
+static int compat_drm_agp_free(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_agp_buffer32_t *req32 = data;
+	struct drm_agp_buffer request;
+
+	request.handle = req32->handle;
+
+	return drm_agp_free_ioctl(dev, (void *)&request, file_priv);
+}
+
+typedef struct drm_agp_binding32 {
+	u32 handle;	/**< From drm_agp_buffer */
+	u32 offset;	/**< In bytes -- will round to page boundary */
+} drm_agp_binding32_t;
+
+static int compat_drm_agp_bind(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_agp_binding32_t *req32 = data;
+	struct drm_agp_binding request;
+
+	request.handle = req32->handle;
+	request.offset = req32->offset;
+
+	return drm_agp_bind_ioctl(dev, (void *)&request, file_priv);
+}
+
+static int compat_drm_agp_unbind(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_agp_binding32_t *req32 = data;
+	struct drm_agp_binding request;
+	
+	request.handle = req32->handle;
+	
+	return drm_agp_unbind_ioctl(dev, (void *)&request, file_priv);
+}
+
+typedef struct drm_scatter_gather32 {
+	u32 size;	/**< In bytes -- will round to page boundary */
+	u32 handle;	/**< Used for mapping / unmapping */
+} drm_scatter_gather32_t;
+
+static int compat_drm_sg_alloc(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_scatter_gather32_t *req32 = data;
+	struct drm_scatter_gather request;
+	int err;
+
+	request.size = (unsigned long)req32->size;
+
+	err = drm_sg_alloc_ioctl(dev, (void *)&request, file_priv);
+	if (err)
+		return err;
+
+	/* XXX not sure about the handle conversion here... */
+	req32->handle = (unsigned long)request.handle >> PAGE_SHIFT;
+
+	return 0;
+}
+
+static int compat_drm_sg_free(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_scatter_gather32_t *req32 = data;
+	struct drm_scatter_gather request;
+
+	request.handle = (unsigned long)req32->handle << PAGE_SHIFT;
+
+	return drm_sg_free(dev, (void *)&request, file_priv);
+}
+
+typedef struct drm_update_draw32 {
+	drm_drawable_t handle;
+	unsigned int type;
+	unsigned int num;
+	/* 64-bit version has a 32-bit pad here */
+	u64 data;	/**< Pointer */
+} __attribute__((packed)) drm_update_draw32_t;
+
+static int compat_drm_update_draw(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_update_draw32_t *update32 = data;
+	struct drm_update_draw request;
+	int err;
+
+	request.handle = update32->handle;
+	request.type = update32->type;
+	request.num = update32->num;
+	request.data = update32->data;
+
+	err = drm_update_draw(dev, (void *)&request, file_priv);
+	return err;
+}
+
+struct drm_wait_vblank_request32 {
+	enum drm_vblank_seq_type type;
+	unsigned int sequence;
+	u32 signal;
+};
+
+struct drm_wait_vblank_reply32 {
+	enum drm_vblank_seq_type type;
+	unsigned int sequence;
+	s32 tval_sec;
+	s32 tval_usec;
+};
+
+typedef union drm_wait_vblank32 {
+	struct drm_wait_vblank_request32 request;
+	struct drm_wait_vblank_reply32 reply;
+} drm_wait_vblank32_t;
+
+static int compat_drm_wait_vblank(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_wait_vblank32_t *req32 = data;
+	union drm_wait_vblank request;
+	int err;
+
+	request.request.type = req32->request.type;
+	request.request.sequence = req32->request.sequence;
+	request.request.signal = req32->request.signal;
+
+	err = drm_wait_vblank(dev, (void *)&request, file_priv);
+	if (err)
+		return err;
+
+	req32->reply.type = request.reply.type;
+	req32->reply.sequence = request.reply.sequence;
+	req32->reply.tval_sec = request.reply.tval_sec;
+	req32->reply.tval_usec = request.reply.tval_usec;
+
+	return 0;
+}
+
+static drm_ioctl_desc_t drm_compat_ioctls[256] = {
+	DRM_IOCTL_DEF(DRM_IOCTL_VERSION32, compat_drm_version, 0),
+	DRM_IOCTL_DEF(DRM_IOCTL_GET_UNIQUE32, compat_drm_getunique, 0),
+	DRM_IOCTL_DEF(DRM_IOCTL_GET_MAP32, compat_drm_getmap, 0),
+	DRM_IOCTL_DEF(DRM_IOCTL_GET_CLIENT32, compat_drm_getclient, 0),
+	DRM_IOCTL_DEF(DRM_IOCTL_GET_STATS32, compat_drm_getstats, 0),
+	DRM_IOCTL_DEF(DRM_IOCTL_SET_UNIQUE32, compat_drm_setunique, DRM_AUTH|DRM_MASTER|DRM_ROOT_ONLY),
+	DRM_IOCTL_DEF(DRM_IOCTL_ADD_MAP32, compat_drm_addmap, DRM_AUTH|DRM_MASTER|DRM_ROOT_ONLY),
+	DRM_IOCTL_DEF(DRM_IOCTL_ADD_BUFS32, compat_drm_addbufs, DRM_AUTH|DRM_MASTER|DRM_ROOT_ONLY),
+	DRM_IOCTL_DEF(DRM_IOCTL_MARK_BUFS32, compat_drm_markbufs, DRM_AUTH|DRM_MASTER|DRM_ROOT_ONLY),
+	DRM_IOCTL_DEF(DRM_IOCTL_INFO_BUFS32, compat_drm_infobufs, DRM_AUTH),
+	DRM_IOCTL_DEF(DRM_IOCTL_MAP_BUFS32, compat_drm_mapbufs, DRM_AUTH),
+	DRM_IOCTL_DEF(DRM_IOCTL_FREE_BUFS32, compat_drm_freebufs, DRM_AUTH),
+	DRM_IOCTL_DEF(DRM_IOCTL_RM_MAP32, compat_drm_rmmap, DRM_AUTH),
+	DRM_IOCTL_DEF(DRM_IOCTL_SET_SAREA_CTX32, compat_drm_setsareactx, DRM_AUTH|DRM_MASTER|DRM_ROOT_ONLY),
+	DRM_IOCTL_DEF(DRM_IOCTL_GET_SAREA_CTX32, compat_drm_getsareactx, DRM_AUTH),
+	DRM_IOCTL_DEF(DRM_IOCTL_RES_CTX32, compat_drm_resctx, DRM_AUTH),
+	DRM_IOCTL_DEF(DRM_IOCTL_DMA32, compat_drm_dma, DRM_AUTH),
+	
+	DRM_IOCTL_DEF(DRM_IOCTL_AGP_ENABLE32, compat_drm_agp_enable, DRM_AUTH|DRM_MASTER|DRM_ROOT_ONLY),
+	DRM_IOCTL_DEF(DRM_IOCTL_AGP_INFO32, compat_drm_agp_info, DRM_AUTH),
+	DRM_IOCTL_DEF(DRM_IOCTL_AGP_ALLOC32, compat_drm_agp_alloc, DRM_AUTH|DRM_MASTER|DRM_ROOT_ONLY),
+	DRM_IOCTL_DEF(DRM_IOCTL_AGP_FREE32, compat_drm_agp_free, DRM_AUTH|DRM_MASTER|DRM_ROOT_ONLY),
+	DRM_IOCTL_DEF(DRM_IOCTL_AGP_BIND32, compat_drm_agp_bind, DRM_AUTH|DRM_MASTER|DRM_ROOT_ONLY),
+	DRM_IOCTL_DEF(DRM_IOCTL_AGP_UNBIND32, compat_drm_agp_unbind, DRM_AUTH|DRM_MASTER|DRM_ROOT_ONLY),
+	
+	DRM_IOCTL_DEF(DRM_IOCTL_SG_ALLOC32, compat_drm_sg_alloc, DRM_AUTH|DRM_MASTER|DRM_ROOT_ONLY),
+	DRM_IOCTL_DEF(DRM_IOCTL_SG_FREE32, compat_drm_sg_free, DRM_AUTH|DRM_MASTER|DRM_ROOT_ONLY),
+	
+	DRM_IOCTL_DEF(DRM_IOCTL_UPDATE_DRAW32, compat_drm_update_draw, DRM_AUTH|DRM_MASTER|DRM_ROOT_ONLY),
+	
+	DRM_IOCTL_DEF(DRM_IOCTL_WAIT_VBLANK32, compat_drm_wait_vblank, DRM_UNLOCKED),
+};
diff --git a/sys/dev/drm2/i915/i915_dma.c b/sys/dev/drm2/i915/i915_dma.c
index f78553e..5644ad6 100644
--- a/sys/dev/drm2/i915/i915_dma.c
+++ b/sys/dev/drm2/i915/i915_dma.c
@@ -1493,6 +1493,10 @@ struct drm_ioctl_desc i915_ioctls[] = {
 	DRM_IOCTL_DEF(DRM_I915_GET_SPRITE_COLORKEY, intel_sprite_get_colorkey, DRM_MASTER|DRM_CONTROL_ALLOW|DRM_UNLOCKED),
 };
 
+#ifdef COMPAT_FREEBSD32
+#include "i915_ioc32.c"
+#endif
+
 struct drm_driver_info i915_driver_info = {
 	.driver_features =   DRIVER_USE_AGP | DRIVER_REQUIRE_AGP |
 	    DRIVER_USE_MTRR | DRIVER_HAVE_IRQ | DRIVER_LOCKLESS_IRQ |
@@ -1516,6 +1520,9 @@ struct drm_driver_info i915_driver_info = {
 	.sysctl_cleanup	= i915_sysctl_cleanup,
 
 	.ioctls		= i915_ioctls,
+#ifdef COMPAT_FREEBSD32
+	.compat_ioctls  = i915_compat_ioctls,
+#endif
 	.max_ioctl	= DRM_ARRAY_SIZE(i915_ioctls),
 
 	.name		= DRIVER_NAME,
diff --git a/sys/dev/drm2/i915/i915_ioc32.c b/sys/dev/drm2/i915/i915_ioc32.c
new file mode 100644
index 0000000..31f81e6
--- /dev/null
+++ b/sys/dev/drm2/i915/i915_ioc32.c
@@ -0,0 +1,124 @@
+/*-
+ * Copyright (C) Paul Mackerras 2005
+ * Copyright (C) Alan Hourihane 2005
+ * All Rights Reserved.
+ *
+ * Permission is hereby granted, free of charge, to any person obtaining a
+ * copy of this software and associated documentation files (the "Software"),
+ * to deal in the Software without restriction, including without limitation
+ * the rights to use, copy, modify, merge, publish, distribute, sublicense,
+ * and/or sell copies of the Software, and to permit persons to whom the
+ * Software is furnished to do so, subject to the following conditions:
+ *
+ * The above copyright notice and this permission notice (including the next
+ * paragraph) shall be included in all copies or substantial portions of the
+ * Software.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+ * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
+ * THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
+ * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
+ * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
+ * IN THE SOFTWARE.
+ *
+ * Authors:
+ *    Paul Mackerras <paulus@samba.org>
+ *    Alan Hourihane <alanh@fairlite.demon.co.uk>
+ */
+
+__FBSDID("$FreeBSD$");
+
+/** @file i915_ioc32.c
+ * 32-bit ioctl compatibility routines for the i915 DRM.
+ */
+
+typedef struct _drm_i915_batchbuffer32 {
+	int start;		/* agp offset */
+	int used;		/* nr bytes in use */
+	int DR1;		/* hw flags for GFX_OP_DRAWRECT_INFO */
+	int DR4;		/* window origin for GFX_OP_DRAWRECT_INFO */
+	int num_cliprects;	/* mulitpass with multiple cliprects? */
+	u32 cliprects;		/* pointer to userspace cliprects */
+} drm_i915_batchbuffer32_t;
+
+static int compat_i915_batchbuffer(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_i915_batchbuffer32_t *batchbuffer32 = data;
+	drm_i915_batchbuffer_t batchbuffer;
+
+	batchbuffer.start = batchbuffer32->start;
+	batchbuffer.used = batchbuffer32->used;
+	batchbuffer.DR1 = batchbuffer32->DR1;
+	batchbuffer.DR4 = batchbuffer32->DR4;
+	batchbuffer.num_cliprects = batchbuffer32->num_cliprects;
+	batchbuffer.cliprects = (void *)(unsigned long)batchbuffer32->cliprects;
+
+	return i915_batchbuffer(dev, (void *)&batchbuffer, file_priv);
+}
+
+typedef struct _drm_i915_cmdbuffer32 {
+	u32 buf;		/* pointer to userspace command buffer */
+	int sz;			/* nr bytes in buf */
+	int DR1;		/* hw flags for GFX_OP_DRAWRECT_INFO */
+	int DR4;		/* window origin for GFX_OP_DRAWRECT_INFO */
+	int num_cliprects;	/* mulitpass with multiple cliprects? */
+	u32 cliprects;		/* pointer to userspace cliprects */
+} drm_i915_cmdbuffer32_t;
+
+static int compat_i915_cmdbuffer(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_i915_cmdbuffer32_t *cmdbuffer32 = data;
+	drm_i915_cmdbuffer_t cmdbuffer;
+
+	cmdbuffer.sz = cmdbuffer32->sz;
+	cmdbuffer.DR1 = cmdbuffer32->DR1;
+	cmdbuffer.DR4 = cmdbuffer32->DR4;
+	cmdbuffer.num_cliprects = cmdbuffer32->num_cliprects;
+	cmdbuffer.cliprects = (void *)(unsigned long)cmdbuffer32->cliprects;
+
+	return i915_cmdbuffer(dev, (void *)&cmdbuffer, file_priv);
+}
+
+typedef struct drm_i915_irq_emit32 {
+	u32 irq_seq;
+} drm_i915_irq_emit32_t;
+
+static int compat_i915_irq_emit(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_i915_irq_emit32_t *req32 = data;
+	drm_i915_irq_emit_t request;
+
+	request.irq_seq = (int *)(unsigned long)req32->irq_seq;
+
+	return i915_irq_emit(dev, (void *)&request, file_priv);
+}
+typedef struct drm_i915_getparam32 {
+	int param;
+	u32 value;
+} drm_i915_getparam32_t;
+
+static int compat_i915_getparam(struct drm_device *dev, void *data, struct drm_file *file_priv)
+{
+	drm_i915_getparam32_t *req32 = data;
+	drm_i915_getparam_t request;
+
+	request.param = req32->param;
+	request.value = (void *)(unsigned long)req32->value;
+
+	return i915_getparam(dev, (void *)&request, file_priv);
+}
+
+typedef struct drm_i915_mem_alloc32 {
+	int region;
+	int alignment;
+	int size;
+	u32 region_offset;	/* offset from start of fb or agp */
+} drm_i915_mem_alloc32_t;
+
+drm_ioctl_desc_t i915_compat_ioctls[DRM_ARRAY_SIZE(i915_ioctls)] = {
+	DRM_IOCTL_DEF(DRM_I915_BATCHBUFFER, compat_i915_batchbuffer, DRM_AUTH),
+	DRM_IOCTL_DEF(DRM_I915_CMDBUFFER, compat_i915_cmdbuffer, DRM_AUTH),
+	DRM_IOCTL_DEF(DRM_I915_GETPARAM, compat_i915_getparam, DRM_AUTH),
+	DRM_IOCTL_DEF(DRM_I915_IRQ_EMIT, compat_i915_irq_emit, DRM_AUTH)
+};
diff --git a/sys/modules/drm2/drm2/Makefile b/sys/modules/drm2/drm2/Makefile
index e73528b..48ea41a 100644
--- a/sys/modules/drm2/drm2/Makefile
+++ b/sys/modules/drm2/drm2/Makefile
@@ -34,6 +34,6 @@ SRCS	= \
 	drm_vm.c
 
 SRCS	+=device_if.h bus_if.h pci_if.h device_if.h iicbus_if.h opt_drm.h \
-	opt_vm.h
+	opt_vm.h opt_compat.h
 
 .include <bsd.kmod.mk>
diff --git a/sys/modules/drm2/i915kms/Makefile b/sys/modules/drm2/i915kms/Makefile
index 611d31f..6b548c6 100644
--- a/sys/modules/drm2/i915kms/Makefile
+++ b/sys/modules/drm2/i915kms/Makefile
@@ -29,6 +29,7 @@ SRCS	= \
 	intel_sdvo.c \
 	intel_sprite.c \
 	intel_tv.c
-SRCS	+= device_if.h bus_if.h pci_if.h iicbus_if.h iicbb_if.h opt_drm.h
+SRCS	+= device_if.h bus_if.h pci_if.h iicbus_if.h iicbb_if.h opt_drm.h \
+	 opt_compat.h
 
 .include <bsd.kmod.mk>